from typing import Dict, Any
import os
import yaml


ConfigDict = Dict[str, Any]


class ConfigFile:
    def __init__(self, filename: str, fail: bool = False) -> None:
        self.filename = filename
        self.fail = fail


class Config:
    DEFAULT_FILES = (
        ConfigFile('config/default.yml', fail=False),
        ConfigFile('config/local.yml', fail=False))

    def __init__(self, config: ConfigDict = None) -> None:
        self._config = config or {}

    def read_all(self, filename: str = None):
        config_files = list(self.DEFAULT_FILES)
        if filename:
            config_files.append(ConfigFile(filename, fail=True))
        for config_file in config_files:
            self.read(config_file.filename, fail=config_file.fail)
        self.load_env()
        return self

    def read(self, filename: str, fail: bool = False):
        try:
            with open(filename) as file:
                self.load(file.read())
        except FileNotFoundError as e:
            if fail:
                raise e
        return self

    def load(self, yaml_contents: str):
        source: ConfigDict = yaml.safe_load(yaml_contents) or {}
        self.merge(source)
        return self

    def load_env(self, variable: str = 'CONFIG'):
        return self.load(os.getenv(variable, ''))

    def __getitem__(self, key: str):
        prop = self._config
        for _key in key.split('.'):
            try:
                prop = prop[_key]
            except KeyError:
                raise KeyError(
                    'Configuration roperty "{}" not defined'.format(key))
        return prop

    def _merge(self, source: ConfigDict, destination: ConfigDict):
        stack = [(source, destination)]

        while stack:
            src, dest = stack.pop()
            for key, value in src.items():
                if isinstance(value, dict):
                    if key not in dest or not isinstance(dest[key], dict):
                        dest[key] = {}
                    stack.append((value, dest[key]))
                    continue
                dest[key] = value

        return destination

    def merge(self, config: ConfigDict):
        self._merge(config, self._config)
        return self

    def __repr__(self):
        return 'Config[_config={}]'.format(self._config)



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='magnet',
    version='1.0.0',
    author='Jerko Steiner',
    author_email='jerko.steiner@gmail.com',
    description=(
        'Hierarchical YAML config reader that reads from multiple files and'
        ' an environment variable'
    ),
    install_requires=[
        'PyYAML>=3.12'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/jeremija/pymagnet',
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



from .config import Config
import pytest
import tempfile


@pytest.fixture
def filename1():
    file = tempfile.NamedTemporaryFile(delete=False)
    yield file.name
    file.close()


@pytest.fixture
def filename2():
    file = tempfile.NamedTemporaryFile(delete=False)
    yield file.name
    file.close()


def test_read_no_file():
    Config().read('/path/to/non/existing/file')


def test_load_env(monkeypatch):
    monkeypatch.setenv('CONFIG', '---\na: b\nc: d\ne:\n  f:\n    - g')
    c = Config().load_env()
    assert c['a'] == 'b'
    assert c['c'] == 'd'
    assert c['e.f'] == ['g']


def test_read_no_file_fail(filename1, filename2):
    with pytest.raises(FileNotFoundError):
        Config().read('/path/to/non/existing/file', fail=True)


def test_read_single_file(filename1):
    with open(filename1, 'w') as file:
        file.write("""---
deeply:
    nested:
        key: test
""")
    config = Config().read(filename1)
    assert config['deeply.nested.key'] == 'test'
    with pytest.raises(KeyError):
        config['invalid.key']


def test_read_multiple_files(filename1, filename2):
    with open(filename1, 'w') as file:
        file.write("""---
keys:
    key1: 1234
    key2:
        a: 1
""")
    with open(filename2, 'w') as file:
        file.write("""---
keys:
    key1:
        value: 'test'
    key2:
        b: 2
""")

    config = Config().read(filename1).read(filename2)
    assert config['keys.key1'] == {'value': 'test'}
    assert config['keys.key2'] == {'a': 1, 'b': 2}


def test_read_all(filename1, monkeypatch):
    with open(filename1, 'w') as file:
        file.write("""---
keys:
    key4: test4
""")

    monkeypatch.setenv('CONFIG', """---
keys:
    key1: test1
    key2: 2
""")

    config = Config().read_all(filename1)
    assert config._config == {
        'keys': {
            'key1': 'test1',
            'key2': 2,
            'key3': 3,
            'key4': 'test4',
        }
    }


def test_read_all_no_filename_no_env():
    config = Config().read_all()
    assert config._config == {
        'keys': {
            'key2': ['a', 'b'],
            'key3': 3,
        }
    }


def test_str():
    assert repr(Config()) == 'Config[_config={}]'
